/*
 * $Id: hashtable.c,v 1.2 2001/08/28 01:38:10 kenta Exp $
 *
 * Copyright 2001 Kenta Cho. All rights reserved.
 */

/**
 * Hashtable utility functions.
 *
 * @version $Revision: 1.2 $
 */
#include <string.h>
#include <stdlib.h>
#include <malloc.h>
#include <stdio.h>

#include "hashtable.h"

void initHash(Hashtable *ht, size_t size, int new) {
  int i;
  ht->elementSize = size;
  for ( i=0 ; i<ALL_ELEMENTS_SIZE ; i++ ) {
    ht->keys[i][0] = '\0';

    if ( !new & ht->valid[i] ) {
      free(ht->elements[i]);
    }
    ht->valid[i] = 0;
  }
}

int hashFunc(char *key) {
  int len = strlen(key), i;
  unsigned int v = 0;
  for ( i=0 ; i<len ; i++ ) {
    v += key[i];
    v += (v << 10);
    v ^= (v >> 6);
  }
  v += (v << 3);
  v ^= (v >> 11);
  v += (v << 15);
  return (v & (KEY_SIZE-1));
}

void* checkByIndex(Hashtable *ht, int h) {
  return ht->elements[h];
}

void* check(Hashtable *ht, char *key, int *valid) {
  int h = hashFunc(key);
  if ( strcmp(ht->keys[h], key) == 0 ) *valid = 1;
  else *valid = 0;
  return ht->elements[h];
}

void* get(Hashtable *ht, char *key, int *valid, int *new) {
  int h = hashFunc(key);
  if ( strcmp(ht->keys[h], key) == 0 ) {
    *valid = 1;
  } else {
    *valid = 0;
    strcpy(ht->keys[h], key);
  }

  if ( !ht->valid[h] ) {
    ht->elements[h] = malloc(ht->elementSize);
    ht->valid[h] = 1;
    *new = 1;
  } else {
    *new = 0;
  }
  return ht->elements[h];
}

// Enumeration functions.

void startEnumeration(Hashtable *ht) {
  ht->enumCnt = -1;
}

int hasMoreElements(Hashtable *ht) {
  ht->enumCnt++;
  while ( ht->enumCnt < ALL_ELEMENTS_SIZE && !(ht->valid[ht->enumCnt]) ) {
    ht->enumCnt++;
  }
  if ( ht->enumCnt < ALL_ELEMENTS_SIZE ) return 1;
  else return 0;
}

void* nextElement(Hashtable *ht) {
  return ht->elements[ht->enumCnt];
}

void deleteElement(Hashtable *ht) {
  ht->keys[ht->enumCnt][0] = '\0';
  ht->valid[ht->enumCnt] = 0;
}

// Save/Load hashtable content to/from file.

void save(Hashtable *ht, FILE *fp) {
  int i, valid;
  putw(ht->elementSize, fp);
  for ( i=0 ; i<ALL_ELEMENTS_SIZE ; i++ ) {
    valid = ht->valid[i];
    putw(valid, fp);
    if ( valid ) {
      fwrite(ht->keys[i], KEY_LENGTH, 1, fp);
      fwrite(ht->elements[i], ht->elementSize, 1, fp);
    }
  }
}

void load(Hashtable *ht, FILE *fp) {
  int i;
  ht->elementSize = getw(fp);
  for ( i=0 ; i<ALL_ELEMENTS_SIZE ; i++ ) {
    if ( (ht->valid[i] = getw(fp)) ) {
      fread(ht->keys[i], KEY_LENGTH, 1, fp);
      ht->elements[i] = malloc(ht->elementSize);
      fread(ht->elements[i], ht->elementSize, 1, fp);
    }
  }
}
